-------------------------------------------------------------------------------------
-- | Namespace Definitions |---------------------------------------------------------
-------------------------------------------------------------------------------------

-- Retrieve the addon name and the namespace table.
local addon, npcg = ...

local functua = npcg.functua
local util = npcg.functua.util

local generator = { }
local data = { }
local tests = { }

-- generators and data do not exist yet
npcg.generators.feature	= generator
npcg.data.feature				= data
npcg.tests.feature			= tests



local function default(replacements)
  local replacements = replacements or { }
  return {
    races       = replacements[     "races"] or { "any" },
		classes			= replacements[   "classes"] or { "any" },
		tribes			= replacements[    "tribes"] or { "any" },
		ages				= replacements[      "ages"] or { "any" },
		genders			= replacements[   "genders"] or { "any" },
		tags				= replacements[      "tags"] or { "any" },
    weight      = 1
  }
end

local features = {
-- a -----------------------------------------------------------------
["arranged %xir hair in braids"] = default({
	
}),
["appears to have an itch"] = default({
	
}),
["appears to be well trained"] = default({
	ages = npcg.data.age.olderThan("teen")
}),
-- b -----------------------------------------------------------------
["blushes easily"] = default({
	
}),
["bites %xir nails"] = default({
	
}),
["bites %xir lower lip"] = default({
	
}),
["bleeds from the left eyebrow"] = default({
	races = npcg.data.race.allRacesExcept(
		"forsaken" ), 
}),
["bleeds from the right eyebrow"] = default({
	races = npcg.data.race.allRacesExcept(
		"forsaken" ), 
}),
-- c -----------------------------------------------------------------
["carries a child of %xir race"] = default({
	ages = npcg.data.age.olderThan(
		"teen")
}),
["carries unsheathed daggers"] = default({
	
}),
["could use a new shirt"] = default({
	
}),
["could use new shoes"] = default({
	
}),
["could use new clothes"] = default({
	
}),
["could use a haircut"] = default({
	
}),
-- d -----------------------------------------------------------------
["definitely needs a bath"] = default({
	
}),
-- e -----------------------------------------------------------------
-- f -----------------------------------------------------------------
["fiddles nervously"] = default({
	
}),
["flexes %xir muscles"] = default({
	
}),
["flashes a smile constantly"] = default({
	
}),
-- g -----------------------------------------------------------------
["grinds %xir teeth most of the time"] = default({
	
}),
["glances to the side every second"] = default({
	
}),
["gestures a lot when speaking"] = default({
	
}),
["grumbles constantly"] = default({
	
}),
-- h -----------------------------------------------------------------
["has a beetle on %xir left shoulder"] = default({
	
}),
["has a crooked nose"] = default({
	
}),
["has a glorious mustache"] = default({
	genders = {
		"male" }, 
	ages = npcg.data.age.olderThan(
		"adult")
}),
["has a large scar across %xir face"] = default({
	
}),
["has a painfully plain look"] = default({
	
}),
["has a plethora of freckles"] = default({
	
}),
["has a scar across %xir throat"] = default({
	
}),
["has a severe lack of eyebrows"] = default({
	
}),
["has a tiny mouth"] = default({
	
}),
["has very small eyes"] = default({
	
}),
["has a very overweight physique"] = default({
	
}),
["has an abnormally large nose"] = default({
	
}),
["has breadcrumbs on %xir face"] = default({
	
}),
["has an eyepatch over %xir right eye"] = default({
	
}),
["has an eyepatch over %xir left eye"] = default({
	
}),
["has blood on %xir clothes"] = default({
	
}),
["has burn scars across half %xir face"] = default({
	
}),
["has canine teeth"] = default({
	
}),
["has gloves on at all times"] = default({
	
}),
["has greasy hair"] = default({
	
}),
["has very frizzy hair"] = default({
	
}),
["has bloodshot eyes"] = default({
	
}),
["has thick eyebrows"] = default({
	
}),
["has bushy eyebrows"] = default({
	
}),
["has a lot of wrinkles"] = default({
	ages = npcg.data.age.olderThan(
		"mature adult")
}),
["has shiny looking skin"] = default({
	
}),
["has pierced eyebrows"] = default({
	
}),
["has tattoos on %xir face"] = default({
	
}),
["has very large ears"] = default({
	
}),
["has very small ears"] = default({
	
}),
["has a lot of pimples"] = default({
	
}),
["has hair coming out of %xir ears"] = default({
	
}),
["has a lot of nosehair"] = default({
	
}),
["has a flat nose"] = default({
	
}),
["has crooked nose"] = default({
	
}),
["has too small legs"] = default({
	
}),
["has very wide hips"] = default({
	
}),
["has hollow cheeks"] = default({
	
}),
["has bulging eyes"] = default({
	
}),
["has tear-filled eyes"] = default({
	
}),
["has a coarse looking face"] = default({
	
}),
["has a dislocated jaw"] = default({
	
}),
["has work-roughened hands"] = default({
	ages = npcg.data.age.olderThan(
		"teen")
}),
["has a mole on %xir left cheek"] = default({
	
}),
["has a mole on %xir right cheek"] = default({
	
}),
["has a mole on %xir nose"] = default({
	
}),
["has mole on %xir upper lip"] = default({
	
}),
["has a large mold covering half %xir face"] = default({
	
}),
["has thin lips"] = default({
	
}),
["has a crooked smile"] = default({
	
}),
["has no front teeth"] = default({
	
}),
["has yellow teeth"] = default({
	
}),
["has needle sharp teeth"] = default({
	
}),
["has rotting teeth"] = default({
	
}),
["has pearl white teeth"] = default({
	
}),
["has a loose front tooth"] = default({
	
}),
["has manboobs"] = default({
	genders = {
		"male" }, 
	ages = npcg.data.age.olderThan(
		"adult")
}),
["has neatly combed hair"] = default({
	
}),
["has no thumb on %xir left hand"] = default({
	
}),
["has no thumb on %xir right hand"] = default({
	
}),
["has pink hair"] = default({
	
}),
["has purple hair"] = default({
	
}),
["has the cutest smile ever"] = default({
	
}),
["has dried tears on %xir cheeks"] = default({
	
}),
["has two different coloured eyes"] = default({
	races = npcg.data.race.allRacesExcept(
		"blood elf",
		"half elf" ), 
}),
["holds %xir daggers in a visible place"] = default({
	
}),
-- i -----------------------------------------------------------------
["is bald"] = default({
	races = npcg.data.race.allRacesExcept(
		"pandaren",
		"tauren" ), 
}),
["is bare feet"] = default({
	
}),
["is moving gracefully"] = default({
	
}),
["is smelling strongly of ale"] = default({
	
}),
["is sunburned"] = default({
	
}),
["is shivering"] = default({
	
}),
["is moving restlessly"] = default({
	
}),
["is cross eyed"] = default({
	
}),
["is having a nose bleed"] = default({
	races = npcg.data.race.allRacesExcept(
		"forsaken" ), 
}),
["is tilted a bit to the left"] = default({
	
}),
["is tilted a bit to the right"] = default({
	
}),
["is dragging %xir feet when walking"] = default({
	
}),
["is speaking painfully slow"] = default({
	
}),
["is speaking fast"] = default({
	
}),
["is drooling"] = default({
	
}),
["is sporting a broken nose"] = default({
	
}),
["is clearly drunk"] = default({
	ages = npcg.data.age.olderThan(
		"teen")
}),
["is cleaning %xir bow"] = default({
	classes = {
		"archer",
    "ranger", 
    "farstrider",
		"dark ranger" },
}),
["is cute in %xir own way"] = default({
	
}),
["is definitely not Blitzcrank"] = default({
	
}), 
["is disguised as Greatfeather Winter"] = default({
	ages = npcg.data.age.olderThan(
		"young adult")
}),
["is humming a song"] = default({
	
}),
["is looking around suspiciously"] = default({
	
}),
["is missing a chunk of %xir right ear"] = default({
	
}),
["is missing %xir left eye"] = default({
	
}),
["is pregant"] = default({
	genders = {
		"female" }, 
	ages = npcg.data.age.olderThan(
		"teen")
}),
["is very short"] = default({
	
}),
["is very tall"] = default({
	
}),
["is waving with a HearthStone deck"] = default({
	
}),
["is wearing a monocle"] = default({
	
}),
-- j -----------------------------------------------------------------
-- k -----------------------------------------------------------------
["keeps %xir hands closed"] = default({
	
}),
["keeps %xir hands in %xir pockets"] = default({
	
}),
["keeps holding onto a piece of blue cloth"] = default({
	
}),
["keeps holding onto a piece of pink cloth"] = default({
	
}),
["keeps holding onto a necklace"] = default({
	
}),
-- l -----------------------------------------------------------------
["laughs audibly"] = default({
	
}),
["looks furious"] = default({
	
}),
["looks sad"] = default({
	
}),
["lacks all %xir teeth"] = default({
	
}),
["looks in pain"] = default({
	
}),
["looks out of place"] = default({
	
}),
["looks uncomfortable"] = default({
	
}),
-- m -----------------------------------------------------------------
["moves %xir feet restlessly"] = default({
	
}),
["moves around restlessly"] = default({
	
}),
["moves gracefully"] = default({
	
}),
["moves slowly"] = default({
	
}),
["moves unsteadily"] = default({
	
}),
["moves with a limp"] = default({
	
}),
["moves with difficulty"] = default({
	
}),
["moves %xir left foot in small circles"] = default({
	
}),
["moves %xir right foot in small circles"] = default({
	
}),
["moves %xir hand to %xir hair constantly"] = default({
	
}),
-- n -----------------------------------------------------------------
["needs heat"] = default({
	
}),
["needs medical attention"] = default({
	
}),
-- o -----------------------------------------------------------------
-- p -----------------------------------------------------------------
["points around when speaking"] = default({
	
}), 
-- q -----------------------------------------------------------------
-- r -----------------------------------------------------------------
["rotates a blacksmith hammer in %xir hand"] = default({
	classes = {
		"blacksmith" },
}), 
-- s -----------------------------------------------------------------
["seems nervous"] = default({
	
}),
["shivers a lot"] = default({
	
}),
["sweats a lot"] = default({
	
}),
["scratches %xir nose a lot"] = default({
	
}),
["scratches %xir arm a lot"] = default({
	
}),
["seems content"] = default({
	
}),
["seems happy"] = default({
	
}),
["seems lovestruck"] = default({
	ages = npcg.data.age.olderThan(
		"teen")
}),
["sniffs sadly"] = default({
	
}),
["sports bright pink goggles"] = default({
	classes = {
		"engineer" }, 
}),
["stares blankly"] = default({
	
}),
-- t -----------------------------------------------------------------
["taps %xir fingers non-stop"] = default({
	
}),
["taps %xir nose a lot"] = default({
	
}),
["twirls %xir hair around a finger"] = default({
	
}),
["taps %xir indext and thumb together"] = default({
	
}),
-- u -----------------------------------------------------------------
-- v -----------------------------------------------------------------
-- w -----------------------------------------------------------------
["wears glasses"] = default({
	
}),
["wears only black clothes"] = default({
	
}),
["wears torn clothes"] = default({
	
}),
["wears a silk cloak"] = default({
	
}),
["wears torn clothes"] = default({
	
}),
["wears a black hat"] = default({
	
}),
["wears a fake beard"] = default({
	
}),
["wears oversized shoes"] = default({
	
}),
["wears a rifle over the shoulder"] = default({
	
}),
["wears a purple skarf"] = default({
	
}),
["wears a black wig"] = default({
	
}),
["wears a blond wig"] = default({
	
}),
["wears a red wig"] = default({
	
}),
["wears a bright yellow cloak"] = default({
	
}),
["wears a monocle"] = default({
	
}),
["wears fancy clothes"] = default({
	
}),
["waves a deck of cards"] = default({
	
}),
-- x -----------------------------------------------------------------
-- y -----------------------------------------------------------------
-- z -----------------------------------------------------------------
} -- End of Feature List.

---------------------------------------------------------------------------------------------------------------------------------------
-- | Functions for Generation and Retrieval | -----------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- Main generator should be CharacterTemplate -> { Trait }
-- This will later change to be CharacterTemplate -> { Trait, weight } or something similar.

-- Part 1: Get potential candidates
-- Part 2: Weed out unfeasable candidates, like those not within context or faction
-- Part 3: Select a random item

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 1 Functions: Generation | -------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions should be of the type:
-- characterTemplate -> zone -> subZone -> { class, classconstraints }

local function dontCheckLocations(characterTemplate, zone, subZone)
  return
    -- Fallback to default.
    features
    -- Fallback in case some extreme breaking happened.
    or { }
end

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 2 Functions: Filtering | --------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions should be of type:
-- function :: characterTemplate -> { class, classconstraints } -> zone -> subzone -> bool

local function meetsTraitConstraints(characterTemplate, option)
	return  functua.table.containsAny(option.races,   { "any", characterTemplate.race   })
    and   functua.table.containsAny(option.ages,    { "any", characterTemplate.age    })
    and   functua.table.containsAny(option.genders, { "any", characterTemplate.gender })
		and   functua.table.containsAny(option.tribes,  { "any", characterTemplate.tribe  })
		and   functua.table.containsAny(option.classes, { "any", characterTemplate.class  })
end

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 3 Functions: Selection | --------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions should be of type:
-- Arbitrary a => { class, a } -> class
-- This might change in the future.

local function randomKeyFrom(options)
  return util.randomFrom(functua.table.keys(options))
end

local function selectUsingWeightedRandom(selectionList)
  return util.weightedRandomFrom(functua.table.keys(selectionList), function(item) return selectionList[item].weight end)
end

---------------------------------------------------------------------------------------------------------------------------------------
--| Combination Function | ------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- createClassGenerator :: Trait a, TraitConstraints b =>
--	(characterTemplate -> zone -> subZone -> { a, b }) -> 
--	(characterTemplate -> { a, b } -> zone -> subzone -> bool) -> 
--	(Arbitrary a => { a, c } -> a)
--	-> (characterTemplate, zone, subZone) -> a

function createFeatureGenerator(generatorF, filterF, selectorF)
  return function(characterTemplate, zone, subZone)
    return selectorF(
      functua.table.filterV( -- Returns KVP, filters on V.
        function(potentialTrait) return filterF(characterTemplate, potentialTrait, zone, subZone) end,  -- Most don't use zone or subzone here.
        generatorF(characterTemplate, zone, subZone)
      )
    )
  end
end

-- This can later be extended.
generateRandomFeature = createFeatureGenerator(dontCheckLocations, meetsTraitConstraints, randomKeyFrom)

--[[

-- satisfiesTraitConstraints :: CharacterTemplate -> FeatureItem -> Bool
local function satisfiesTraitConstraints(characterTemplate, potentialFeature)
	return (
		functua.contains(characterTemplate:getRace(),		potentialFeature.races)		or 
		functua.contains("any", potentialFeature.races)
	) and (
		functua.contains(characterTemplate:getClass(),	potentialFeature.classes)	or
		functua.contains("any", potentialFeature.classes)
	) and (
		functua.contains(characterTemplate:getTribe(),	potentialFeature.tribes)	or
		functua.contains("any", potentialFeature.tribes)
	) and (
		functua.contains(characterTemplate:getGender(),	potentialFeature.genders)	or
		functua.contains("any", potentialFeature.genders)
	) and (
		functua.contains(characterTemplate:getAge(),		potentialFeature.ages)		or
		functua.contains("any", potentialFeature.ages)
	)
end

-- generateRandomFeature :: (Race a, Class a, Gender a, Age a) => a -> Feature
function generateRandomFeature(characterTemplate)
  local d = date("*t")
  if d.month == 4 and d.day == 1 then return "is carrying a Left Shark"
  else
		return
			util.randomFrom(
				functua.filterKeysByValue(
					function(val) return satisfiesTraitConstraints(characterTemplate, val) end,
					features
				)
			) or "Error: No options found!"
	end -- April fool end
end
]]
-- End of File.