-------------------------------------------------------------------------------------
-- | Namespace Definitions |---------------------------------------------------------
-------------------------------------------------------------------------------------

-- Retrieve the addon name and the namespace table.
local addon, npcg = ...

local functua = npcg.functua
local util 		= npcg.functua.util

local generator = { }
local data 			= { }
local tests 		= { }

-- generators and data do not exist yet
npcg.generators.tribe	= generator
npcg.data.tribe				= data
npcg.tests.tribe			= tests

---------------------------------------------------------------------------------------------------------------------------------------
-- | Raw Data | -----------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- Data Table, this will be loaded in externally. Do not use this directly, use the functions instead.
local defaultTable = {
	[ "death knight"  ] = {
		[ "draenei"     ] = {
			weight = 1,
			traitTypes = { "alliance", "retail", }, },
		[ "dwarf"       ] = {
			weight = 2,
			traitTypes = { "alliance", "retail", }, },
		[ "gnome"       ] = {
			weight = 1,
			traitTypes = { "alliance", "retail",}, },
		[ "goblin"      ] = {
			weight = 1,
			traitTypes = { "horde", "retail", }, },
		[ "high elf"    ] = {
			weight = 4,
			traitTypes = { "neutral", "retail", }, },
		[ "human"       ] = {
			weight = 5,
			traitTypes = { "alliance", "retail", }, },
		[ "night elf"   ] = {
			weight = 1,
			traitTypes = { "alliance", "retail", }, },
		[ "orc"         ] = {
			weight = 1,
			traitTypes = { "horde", "retail", }, },
		[ "tauren"      ] = {
			weight = 1,
			traitTypes = { "horde", "retail", }, },
		[ "troll"       ] = {
			weight = 2,
			traitTypes = { "horde", "retail", }, },
		[ "worgen"      ] = {
			weight = 1,
			traitTypes = { "alliance", "retail", } }, },
	[ "draenei"       ] = {
		[ ""            ] = {
			weight = 6,
			traitTypes = { "alliance", "retail", }, },
		[ "lightforged" ] = {
			weight = 1,
			traitTypes = { "alliance", "retail", } }, },
	[ "dwarf"         ] = {
		[ "dark iron"   ] = {
			weight = 1,
			traitTypes = { "alliance", "retail", }, },
		[ "wildhammer"  ] = {
			weight = 1,
			traitTypes = { "alliance", "retail", }, },
		[ "bronzebeard" ] = {
			weight = 1,
			traitTypes = { "alliance", "classic", "retail", }, },
		[ "frostborn"   ] = {
			weight = 1,
			traitTypes = { "alliance", "retail", }, },
		[ "stonefist"   ] = {
			weight = 1,
			traitTypes = { "alliance", "retail", }, },
		[ "stormpike"   ] = {
			weight = 1,
			traitTypes = { "alliance", "classic", "retail", } }, },
	[ "forsaken"      ] = {
		[ "blood elf"   ] = {
			weight = 1,
			traitTypes = { "horde", "retail", }, },
		[ "high elf"    ] = {
			weight = 2,
			traitTypes = { "horde", "classic", "retail", }, },
		[ "human"       ] = {
			weight = 7,
			traitTypes = { "horde", "classic", "retail", } }, },
  [ "gnome"         ] = {
    [ "gnomeregan"  ] = {
      weight = 20,
      traitTypes = { "alliance", "classic", "retail", }, },
    [ "mechagon"    ] = {
      weight = 1,
      traitTypes = { "alliance", "retail", } }, },
	[	"goblin"				] = {
		[	"bilgewater"	] = {
			weight = 8,
			traitTypes = { "horde", "retail", }, },
		[ "steamweedle" ] = {
			weight = 1,
			traitTypes = { "neutral", "classic", "retail", } }, },
	[ "human"					] = {
		[ "dalaran" 		] = {
			weight = 5,
			traitTypes = { "alliance", "retail", }, },
		[ "lordaeron"		] = {
			weight = 1,
			traitTypes = { "alliance", "classic", "retail", }, },
		[ "kul'tiras"   ] = {
			weight = 3,
			traitTypes = { "alliance", "retail", }, },
		[ "stormwind"		] = {
			weight = 9,
			traitTypes = { "alliance", "classic", "retail", }, },
		[ "stromgarde"  ] = {
			weight = 1,
			traitTypes = { "alliance", "classic", "retail", }, },
		[ "theramore"		] = {
			weight = 1,
			traitTypes = { "alliance", "classic", "retail", } }, },
	[ "illidari"			] = {
		[ "blood elf"		] = {
			weight = 1,
			traitTypes = { "horde", "retail", }, },
		[ "night elf"		] = {
			weight = 2,
			traitTypes = { "alliance", "retail", } }, },
	[ "night elf"     ] = {
		[ "kaldorei"    ] = {
			weight = 8,
			traitTypes = { "alliance", "classic", "retail", }, },
		[ "highborne"   ] = {
			weight = 1,
			traitTypes = { "alliance", "retail", } }, },
	[ "orc"           ] = {
		[ "fel"         ] = {
			weight = 4,
			traitTypes = { "horde", "classic", "retail", }, },
		[ "brown"       ] = {
			weight = 1,
			traitTypes = { "horde", "retail", }, },
		[ "mag'har"			] = {
			weight = 1,
			traitTypes = { "horde", "retail", } }, },
	[ "pandaren"      ] = {
		[ "huojin"      ] = {
			weight = 20,
			traitTypes = { "horde", "retail", }, },
		[ "tushui"      ] = {
			weight = 20,
			traitTypes = { "alliance", "retail", }, },
		[ "independant" ] = {
			weight = 1,
			traitTypes = { "neutral", "retail", } }, },
	[ "tauren"        ] = {
		[ "bloodhoof"   ] = {
			weight = 1,
			traitTypes = { "horde", "classic", "retail", }, },
		[ "dawnstrider" ] = {
			weight = 1,
			traitTypes = { "horde", "classic", "retail", }, },
		[ "grimtotem"   ] = {
			weight = 1,
			traitTypes = { "horde", "classic", "retail", }, },
		[ "mistrunner"  ] = {
			weight = 1,
			traitTypes = { "horde", "classic", "retail", }, },
		[ "ragetotem"   ] = {
			weight = 1,
			traitTypes = { "horde", "classic", "retail", }, },
		[ "runetotem"   ] = {
			weight = 1,
			traitTypes = { "horde", "classic", "retail", }, },
		[ "skychaser"   ] = {
			weight = 1,
			traitTypes = { "horde", "classic", "retail", }, },
		[ "stonehoof"   ] = {
			weight = 1,
			traitTypes = { "horde", "classic", "retail", }, },
		[ "thunderhorn" ] = {
			weight = 1,
			traitTypes = { "horde", "classic", "retail", }, },
		[ "wildmane"    ] = {
			weight = 1,
			traitTypes = { "horde", "classic", "retail", }, },
		[ "winterhoof"  ] = {
			weight = 1,
			traitTypes = { "horde", "classic", "retail", }, },
		[ "clawhoof"    ] = {
			weight = 1,
			traitTypes = { "horde", "classic", "retail", }, },
		[ "highmountain"] = {
			weight = 1,
			traitTypes = { "horde", "retail", } }, },
	[ "troll"         ] = {
		[ "amani"       ] = {
			weight = 1,
			traitTypes = { "horde", "retail", }, },
		[ "darkspear"   ] = {
			weight = 6,
			traitTypes = { "horde", "classic", "retail", }, },
		[ "farraki"     ] = {
			weight = 1,
			traitTypes = { "horde", "retail", }, },
		[ "gurubashi"   ] = {
			weight = 1,
			traitTypes = { "horde", "retail", }, },
		[ "zandalari"   ] = {
			weight = 4,
			traitTypes = { "horde", "retail", }, }, },
}

-- Contextual meaning of "Tribe" for the races. This is also the title displayed in the addon. Empty strings mean Tribe has no context for the race.
local context	= { 
  ["blood elf"    ] = "",
  ["death knight" ] = "Died as",
  ["draenei"      ] = "Type",
  ["dwarf"        ] = "Clan",
  ["gnome"        ] = "Origin",
  ["goblin"       ] = "Cartel",
  ["half elf"     ] = "",
  ["half orc"     ] = "",
  ["high elf"     ] = "",
  ["human"        ] = "Origin",
	["illidari"			] = "Past",
  ["night elf"    ] = "Type",
  ["orc"          ] = "Clan",
  ["pandaren"     ] = "Faction",
  ["tauren"       ] = "Clan",
  ["troll"        ] = "Tribe",
  ["forsaken"     ] = "Died as",
  ["worgen"       ] = "",
	["a"						] = "Test",
}

---------------------------------------------------------------------------------------------------------------------------------------
-- | Data Modifiers | -----------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions allow for customization of defaultTable

-- default :: { Race, { Tribe, TribeConstraints } } -> { Race, { Tribe, TribeConstraints } }
-- Returns the default tribe distribution template altered by entries within the given table.
-- Note: If replacements contains anything not present in the default setting, it will not be returned.
local function default(replacements)
	local replacements = replacements or { }
	
	return functua.table.mapKV(
		function(race, tribes)	-- takes in the key and value
			return functua.table.mapKV(
				function(tribeKey, tribeValue) 
					return ((replacements[race] or { })[tribeKey]) or functua.table.clone(tribeValue) end,
				tribes)
		end,
		defaultTable)
end

-- defaultFromWeightTable :: { Race, { Tribe, Int } } -> { Race, { Tribe, TribeConstraints } }
-- Returns the default tribe distribution template altered by a weight table. The default weight is 1.
-- Note: If replacements contains anything not present in the default setting, it will not be returned.
local function defaultFromWeightTable(replacements)
	local output = default({ })
	
	for race,tribes in pairs(replacements) do
		for tribe, value in pairs(tribes) do
			if output[race] ~= nil and output[race][tribe] ~= nil then
				output[race][tribe].weight = value end
		end
	end
	return output
end

-- neutral :: { Race, { Tribe, TribeConstraints } } -> { Race, { Tribe, TribeConstraints } }
-- Returns a tribe distribution where all tribes are given a neutral tribeType.
local function neutral(entries)
	return functua.table.map(
		function(xs) return -- xs :: { Tribe, TribeConstraints }
			functua.table.map(
				function(x)
					local output = functua.table.map(functua.id, x) -- Clone x.
					output.traitTypes = functua.list.insert(output.traitTypes, "neutral")
					return output
				end,
				xs) end,
		entries)
end

---------------------------------------------------------------------------------------------------------------------------------------
-- | Active Data Templates | ----------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- The default tribe distribution template, this is called when there are no specifics for zones and/or subzones.
local tribes = default({ })

---------------------------------------------------------------------------------------------------------------------------------------
-- | Zone-based Generation | ----------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

local zonesToTribes = {

-- Alliance Capitals
["Stormwind City"			] = defaultFromWeightTable({
	["human"				] = {
		["stormwind"			] = 100, }, }),

-- Neutral
["Dalaran"        		] = neutral(defaultFromWeightTable({
	["human"				] = {
		["dalaran"  			] = 100, }, }) ),

-- Warfronts
["Arathi Highlands"		] = neutral(default({ })),
["Darkshore"					] = neutral(default({ })),

-- Battle for Azeroth

-- Kul'tiras
["Boralus" 						] = defaultFromWeightTable({
	["human"				] = {
		["kul'tiras"			] = 100, }, }),
["Drustvar"						] = defaultFromWeightTable({
	["human"				] = {
		["kul'tiras"			] = 100, }, }),
["Stormsong Valley"		] = defaultFromWeightTable({
	["human"				] = {
		["kul'tiras"			] = 100, }, }),
["Tiragarde Sound"		] = defaultFromWeightTable({
	["human"				] = {
		["kul'tiras"			] = 100, }, }),

-- Zandalar
["Dazar'alor"					] = defaultFromWeightTable({ -- Dazar'alor is only a select number of subzones.
	["troll"				] = {
		["zandalari"			] = 20, }, }),
["Nazmir"							] = defaultFromWeightTable({
	["troll"				] = {
		["zandalari"			] = 20, }, }),
["Vol'dun"						] = defaultFromWeightTable({
	["troll"				] = {
		["zandalari"			] = 20, }, }),
["Zuldazar"						] = defaultFromWeightTable({
	["troll"				] = {
		["zandalari"			] = 20, }, }),

}

---------------------------------------------------------------------------------------------------------------------------------------
-- | Subzone-based Generation | -------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

local subzonesToTribes = {
-- Order Halls

["Mardum, the Shattered Abyss"	]	= {
	["The Fel Hammer"] = neutral(default({ })), },
	
-- Booty Bay
["The Cape of Stranglethorn"		] = {
	["Booty Bay"							] = neutral(defaultFromWeightTable({
		["goblin"			] = {
			["bilgewater"		] = 1,
			["steamweedle"	] = 9, }, })),
	["The Old Port Authority" ] = neutral(defaultFromWeightTable({
		["goblin"			] = {
			["bilgewater"		] = 1,
			["steamweedle"	] = 9, }, })), },
["Booty Bay"										] = {
	["The Salty Sailor Tavern"] = neutral(defaultFromWeightTable({
		["goblin"			] = {
			["bilgewater"		] = 1,
			["steamweedle"	] = 9, }, })), },

}

---------------------------------------------------------------------------------------------------------------------------------------
--| Queries | -------------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- allRacesExcept :: Tribe -> ... -> { Tribe }
function data.allTribesExcept(...)
	return functua.table.without(functua.table.concatKeys(tribes), {...})
end

-- getTribeContext :: Race -> string
function data.getTribeContext(race)
	return context[race] or ""
end

---------------------------------------------------------------------------------------------------------------------------------------
-- | Functions for Generation and Retrieval | -----------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- Main generator should be CharacterTemplate -> { Trait }
-- This will later change to be CharacterTemplate -> { Trait, weight } or something similar.

-- Part 1: Get potential candidates
-- Part 2: Weed out unfeasable candidates, like those not within context or faction
-- Part 3: Select a random item

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 1 Functions: Generation | -------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- TribeDetails a => CharacterTemplate -> ZoneString -> SubZoneString -> { Tribe, a }

local function checkAllLocations(characterTemplate, zone, subZone)
  local race = characterTemplate.race
	local output =
	    -- Test whether the SubZone has values.
    ((subzonesToTribes[zone] or { })[subZone] or { })[race]
    -- Test whether the Zone has values.
    or (zonesToTribes[zone] or { })[race]
    -- Fallback to default.
    or tribes[race]
    -- Fallback in case some extreme breaking happened.
    or { [""] = { weight=1,traitTypes= { "neutral", "weird", "classic" } } }
	
	-- Duplicate..?
	
	return 
    -- Test whether the SubZone has values.
    ((subzonesToTribes[zone] or { })[subZone] or { })[race]
    -- Test whether the Zone has values.
    or (zonesToTribes[zone] or { })[race]
    -- Fallback to default.
    or tribes[race]
    -- Fallback in case some extreme breaking happened.
    or { [""] = { weight=1,traitTypes= { "neutral", "weird", "classic" } } }
end

local function checkZonesOnly(characterTemplate, zone)
  local race = characterTemplate.race
  return
    -- Test whether the Zone has values.
    (zonesToTribes[zone] or { })[race]
    -- Fallback to default.
    or tribes[race]
    -- Fallback in case some extreme breaking happened.
    or { [""] = { weight=1,traitTypes= { "neutral", "weird", "classic" } } }
end

local function dontCheckLocations(characterTemplate)
  local race = characterTemplate.race
  return
    -- Fallback to default.
    tribes[race]
    -- Fallback in case some extreme breaking happened.
    or { [""] = { weight=1,traitTypes= { "neutral", "weird", "classic" } } }
end

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 2 Functions: Filtering | --------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions should be of type:
-- characterTemplate -> tribe -> zone -> subzone -> bool

local function isClassicTribe(characterTemplate, tribe)
  return functua.list.contains("classic", tribe.traitTypes)
end

local function isRetailTribe(characterTemplate, tribe)
  return functua.list.contains("retail", tribe.traitTypes)
end

local function isNextExpansionTribe(characterTemplate, tribe)
  return functua.list.contains("future", tribe.traitTypes)
end

local function isTribeOfFaction(characterTemplate, tribe)
	return
    functua.table.any(
      function(x) return (x == string.lower(UnitFactionGroup("player")) or x == "neutral") end, 
      tribe.traitTypes)
end

local function isTribeWithinContext(characterTemplate, tribe, zone, subZone)
	return isTribeOfFaction(characterTemplate, tribe)
end

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 3 Functions: Selection | --------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions should be of type:
-- { tribe, stuff } -> tribe
-- This might change in the future.

local function selectUsingWeightedRandom(selectionList)
  return util.weightedRandomFrom(functua.table.keys(selectionList), function(trait) return selectionList[trait].weight end)
end

---------------------------------------------------------------------------------------------------------------------------------------
--| Combination Function | ------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

local function createTribeGenerator(generatorF, filterF, selectorF)
  return function(characterTemplate, zone, subZone)
    return selectorF(
      functua.table.filterV(
        function(trait) return filterF(characterTemplate, trait, zone, subZone) end, 
        generatorF(characterTemplate, zone, subZone))) or ""
  end
end

-- Returns the generatorF argument for createRaceGenerator based on the current options set in NPCG_Options.locationSettings
local function getGeneratorFunction()
  if     not NPCG_Options.locationSettings.zones then
    return dontCheckLocations
  elseif not NPCG_Options.locationSettings.subZones then
    return checkZonesOnly
  else
    return checkAllLocations
  end
end

-- Returns the filterF argument for createRaceGenerator based on the current options set in NPCG_Options.factionSettings and the Game Level.
local function getFilterFunction()
  local output = functua.const(true)
  
  -- Check faction settings.
  if      not NPCG_Options.factionSettings.matching then
    output = functua.const(true)
  elseif  not NPCG_Options.factionSettings.context then
    output = isTribeOfFaction
  else
    output = isTribeWithinContext
  end
  
  -- Check Expansion Level.
  return functua.bothL(output, util.eitherOrByExpansionLevel(isClassicTribe, isRetailTribe, isNextExpansionTribe))
end

-- Set the Tribe Generator based on settings.
-- Must be called before calling "generateRandomTribe".
-- Depends on NPCG_Options.
function generator.initialize()
  generateRandomTribe = function() return "Error: generateRandomTribe is not set" end
  
  -- Arguments for createTraitGenerator
  local generatorF  = getGeneratorFunction()
  local filterF     = getFilterFunction()
  local selectorF   = selectUsingWeightedRandom
  
  generateRandomTribe = createTribeGenerator(generatorF, filterF, selectorF)
end

---------------------------------------------------------------------------------------------------------------------------------------
--| Exposing Functions and Data | -----------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

data.defaultTable     = tribes
data.zoneDataTable    = zonesToTribes
data.subZoneDataTable = subzonesToTribes

---------------------------------------------------------------------------------------------------------------------------------------
--| Testing Functions | ---------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------


-- End of File